/*
	Render_to_Renderfarm.jsx

	Other necessary files include:
		Renderfarm_Settings.ars
		Renderfarm_Output_Module.aom
		touchFiles.sh
		com.graphics.AFXRenderfarmWatch.plist

	Version 1.0 March 2009 by Adam Katz

	I know this works on CS3, I don't know about other versions.

	
	So this is a script to automatically set up a renderfarm for After Effects, including making any folders necessary.


	1: All files that are used in your project must be located on a computer that all render machines can see.

	2: Add the Renderfarm_Settings.ars and Renderfarm_Output_Module.aom to your After Effects render templates

	3: Add compositions to the render queue as normal, and choose where they will render.

	4: Edit the variable renderfarm_location defined below. It is where the renderfarm files will be located, and it should be located on a server that all machines can access.

	5: OPTIONAL -- This step depends on if you are running at Mac OSX 10.4. See Note 1 below for further details. 
		Copy touchFiles.sh and com.graphics.AFXRenderFarmWatch.plist to /Library/LaunchDaemons on all machines that will be rendering

	6: Run the script.

	7: A dialog will appear asking if your render folder(s) be seen by all computers?
		If the location you will render to is accessible to all computers that will be rendering, click yes. If not, the script will make the folders necessary for rendering to occur, and all final renders
		will be located in a folder "Renders" in your renderfarm directory

	8: After you run the script, take any computers you want to render the project, and choose File-->Watch Folder, and select the After Effects renderfarm folder you specified below.




	The script will take any composition you have queued to render, and precompose all layers in it. It will then set this to render as an image sequence with "Skip existing files" checked.
	It will also get set to Import and Replace Usage as a Post-Render Action.
	This will then be in the original composition, which will render with your original output settings.

	Within the renderfarm directory, a directory will be made for each project you send.
	Within this folder, a copy of your project will be saved, along with a render control file, and time file showing what time the project was sent.
	There will also be a "renders" folder which will have a folder for each item that was in the render queue.

*/




/*NOTES

	NOTE 1

	This applies to Mac OSX 10.4, I do not know if it occurs on another version of Mac OSX, or on PCs.

	An issue occurs if the following happens:
	A file is sent to the renderfarm
	The file is rendered.
	The file or folder is then deleted.
	The file is sent again to the renderfarm.

	It seems that the interal updating of files on Mac OSX does not know that the file exists anymore. You can see this by (in a Terminal) running "ls" on the project's directory in the renderfarm folder.
	This will display the names of the After Effects Project file, and the rendefarm control file. However, if you run "ls -l", you will get an error saying the files are not found. This leads to the
	After Effect Render Engine not knowing that new files are there, so they will never start rendering them.

	To fix this, a couple things are needed.

	First, the command "touch" must be run on all folder inside the Renderfarm directory. Also, it must be run on all files within all the folders.
	Example:

	touch renderfarm_directory / *
	touch renderfarm_directory /* /*   <---Note: There should not be a space between the 1st "*" and the 2nd "/"  (javascript comment syntax is fun)

	These 2 commands should be placed in a bash script. In my example, named touchFiles.sh
	Wherever the script is located, needs to be placed in the machine's path. As far as I can tell, this script needs to be on every computer that will be rendering.

	Next, there must be a LaunchDaemon running. 
	The plist file should be placed in "Main Hard Drive / Library / LaunchDaemons".
	This will tell the computer to run the touchFiles script, whenever something is added to the After Effects Render folder.
	To accomplish this, a temporary file is added to the renderfarm folder when this script begins, and deleted at the end of the script.

	So, to summarize Note 1:
	This renderfarm script adds and deletes a temporary file to the renderfarm folder
	A file with the time and date when the After Effects project was added gets created. (The timestamp on the files will later change, so this gives us a reference of when the project was sent to the renderfarm)
	The LaunchDaemon notices the change in the folder
	touchFiles.sh gets run
	All the renderfarm folders and files get their time stamp modified
	After Effects sees the files and render goes along smoothly
	Go have a drink and a bite of food

	NOTE 2
	The script will not precompose for still frames. A waste of time and space.

*/
{
	var safeToRunScript = false;

	var securitySetting = app.preferences.getPrefAsLong("Main Pref Section", "Pref_SCRIPTING_FILE_NETWORK_SECURITY");
	if (!securitySetting)
		alert(" This script requires the scripting security preference to be set.\n" +
		"Go to the \"General\" panel of the application preferences and make sure " + 
	        "\"Allow Scripts to Write Files and Access Network\" is checked.");
	else
		safeToRunScript = true;
	
	var numRenderQueueItems = app.project.renderQueue.numItems;
	
	if (!numRenderQueueItems && safeToRunScript) {
		alert("There must be at least 1 item in the render queue in order to run the script.");
		safeToRunScript = false;
	}
	
	if (numRenderQueueItems && safeToRunScript) {
		var numRenderItemsQueued = 0
		for (i = 1; i <= app.project.renderQueue.numItems; i++) {
			if (app.project.renderQueue.item(i).status == RQItemStatus.QUEUED) {
					numRenderItemsQueued = 1;
					break;
			} 
		}
		if (!numRenderItemsQueued) {
			safeToRunScript = false;
			alert("There must be at least 1 render item queued in order to run the script.");
		}
	}
	
	

	if (safeToRunScript) {

		win = new Window('dialog', "Let's Set Up a Renderfarm!!!", "x:350,y:350,width:350,height:175");
		win.grp1 = win.add('group', "x:0,y:0,width:400,height:400");
	
		win.grp1.infoTxt = win.grp1.add('statictext', "x:15,y:15,width:300,height:20", "Can your render folder(s) be seen by all computers?");
		win.grp1.renderfolderBtnYes = win.grp1.add('radiobutton', "x:15,y:35,width:50,height:20", 'Yes');
		win.grp1.renderfolderBtnNo = win.grp1.add('radiobutton', "x:65,y:35,width:100,height:20", 'No');
		win.grp1.renderfolderBtnNo.value = true;
	
		win.grp1.runBtn = win.grp1.add('button', "x:15,y:80,width:300,height:30",	'Click to set up renderfarm file', {name:'ok'});
		win.grp1.runBtn .onClick = makeRenderfarmFile;
		win.grp1.cancelBtn = win.grp1.add('button', "x:15,y:120,width:100,height:30", 'Cancel', {name:'cancel'});
	
		win.show();
	}
}


function makeRenderfarmFile() {
	
	win.close();
	
	var render_Queue  = app.project.renderQueue;
	var output_modules = render_Queue.outputModules;

	var num_Render_Items = render_Queue.numItems;

	if (num_Render_Items >= 1) {
	
		app.beginUndoGroup("Render to Renderfarm");
		
		//This is the location where the renderfarm will be located.
		//You must chage this variable to something on the server that all computer can access
		var renderfarm_location = "/Volumes/Graphics_Projects_2/_After_Effects_Renderfarm/";


		var render_to_accessible_server=0;	//0=final render must change to a folder accessible by all machines
							//1=final render already accessible, do not change
		
		render_to_accessible_server = (win.grp1.renderfolderBtnYes.value == true) ? 1 : 0;

		
		// These next few lines are necessary to resolve a bug that occurs at least with Mac 10.4. It may affect other operating systems as well.
		// see: Note 1
		// We need to make the Renderfarm machines update their timestamps of all files. This is done through a Launch daemon. This daemon will update time stamps if someting
		// is changed in the top most renderfarm folder ("/Volumes/Graphics Peojects/_After_Effects_Renderfarm"). So, we add an empty file here, and delete it at the end of the script.
		// This will actually cause the timestamp to update twice (addinng and removing the file), but there's no need to have the file always there.
		var temp_Renderfarm_File = new File(renderfarm_location + "temp_Renderfarm_File");
		temp_Renderfarm_File.open("w");
		temp_Renderfarm_File.close();
			
		var version_number = "01";	//version number of the After Effects Render File  CAN NOT be more than 99 versions
		var filename = getFilename(app.project.file.name);
		filename = convertDashToUnderscore(filename);
		if (filename == "temp_Renderfarm_File") //don't have a folder named the same as a the temp_Renderfarm_File which will be deleted at the end of the script
			filename = "temp_Renderfarm_f";
		var human_readable_filename = unescape_spaces(filename);
		var new_project_folder = renderfarm_location + filename + "/";
		var filename_truncated = filename.substring(0,13);

		renderfarm_filename = filename_truncated + "_rfarm" + "-" + version_number;
		var full_filename = new_project_folder + renderfarm_filename + ".aep";
			
		// see: Note 1
		// also, we have to rebuild our "renderfarm_filename" with the new version number, so the new Render Control File is correct
		full_filename_array = check_File_Version(full_filename, 0);
		full_filename = full_filename_array[0];
		version_number = full_filename_array[1];
		if (version_number != "0")
			renderfarm_filename = filename_truncated + "_rfarm" + "-" + version_number;
		// quit the script if there are too many file versions on the server
		if (full_filename == 0) {
			alert("There are too many versions of this file on the renderfarm. There can not be more than 99.\nDelete some and resend the project.");
			return 0;
		}

			
		var new_render_folder = new_project_folder + "renders/";

		var renderfarm_render_folder = renderfarm_location + "Renders/" + filename + "/";

		
		
		//make a folder to hold this project
		projectFolder = new Folder(new_project_folder);
		projectFolder.create();

		//make a folder to render the image sequence to
		renderFolder = new Folder(new_render_folder);
		renderFolder.create();

		//make a folder to render the final output to
		if (!render_to_accessible_server) {
			renderfarmRenderFolder = new Folder(renderfarm_render_folder);
			renderfarmRenderFolder.create();
		}

		var compsToDuplicate = Array();
		
		var num_Queued_Items = 0;  //so we can throw an alert if there are no items set to render
		
		for (var i = 1; i <= num_Render_Items; i++) {

			var curItem = render_Queue.item(i);			
			
			if (curItem.status == RQItemStatus.QUEUED) {

				var curComp = curItem.comp;
				var framesToRender = curItem.timeSpanDuration * curComp.frameRate	//NOTE 2: we don't want to precomp still frames, a waste of time
				framesToRender = Math.round(framesToRender);
			
				if (framesToRender > 1)
					compsToDuplicate[compsToDuplicate.length] = curItem;
							
				num_Queued_Items++;
				
				curCompName = curComp.name;
				
				if (framesToRender > 1) {

					var new_comp_render_folder = new_render_folder + curCompName + "/";
					//make a folder to render the image sequence to
					comp_renderFolder = new Folder(new_comp_render_folder);
					comp_renderFolder.create();

					//build a precomp to render to an image sequece
					var allLayers = new Array();
					var curCompLayers = curComp.layers;
					var numLayers = curCompLayers.length;
					//make an array to hold all the layers to add to the precomp
					for (var j=1;j<=numLayers;j++) {
						allLayers[j-1] = j;
					}
					newCompName = curCompName + "_imagesequence";
					newCompName = newCompName.substring(0,31);
					imageSequce_precomps = curCompLayers.precompose(allLayers, newCompName, true);
			
					var workAreaStart = curComp.workAreaStart;
					var curCompPrecompLayer = curComp.layers[1]; //assume only 1 layer, since everything has been put into a precomp
					curCompPrecompLayer.startTime = workAreaStart;
							
					var imageSequence_render = render_Queue.items.add(imageSequce_precomps);
					imageSequence_render.applyTemplate("Renderfarm Settings");
					imageSequence_render.outputModule(1).applyTemplate("Renderfarm Output Module");
					var oldLocation = imageSequence_render.outputModule(1).file;
					var new_output_name = comp_renderFolder.toString() + "/" + oldLocation.name;
					imageSequence_render.outputModule(1).file = new File(new_output_name);
				}
							
				if (!render_to_accessible_server) {
					for (var k=1; k<=curItem.numOutputModules; k++) {
						var curItemOM = curItem.outputModule(k);
						var curItemOldLocation = curItemOM.file;
						var curItemNewLocation = renderfarmRenderFolder.toString() + "/" + curItemOldLocation.name;
						curItemOM.file = new File(curItemNewLocation);
					}
				}
			
				if (framesToRender > 1) {
					//having duplicate here, means that 1 image sequence gets rendered, then gets made into a movie file, then next image sequence...
					var duplicateItem = curItem.duplicate();
				}
			}
		}

		for (var k=0;k<compsToDuplicate.length;k++) {
			compsToDuplicate[k].remove();
		}

		if (!num_Queued_Items)
			alert("You must have at least one item set to render in the Render Queue");
			
		else {

			//save our file in the renderfarm folder
			new_render_file = new File(full_filename);
			app.project.save(new_render_file);
	
			make_RenderControlFile(new_project_folder + renderfarm_filename);
			
			var timestamp_file = new File(new_project_folder + renderfarm_filename + "-" + get_formatted_date());
			timestamp_file.open("w");
			timestamp_file.close();
			
	
			app.project.close(CloseOptions.DO_NOT_SAVE_CHANGES);
		
			//output a message to the user telling them their file is saved, and optionally, the directory where there render will be
			if (!render_to_accessible_server)
				alert("File sent to Renderfarm\n\nRenders will be saved to\n "+ unescape_spaces(renderfarm_location) + human_readable_filename);
			else
				alert("File sent to Renderfarm\n\nRenders will be saved to the folder where you set them");

			temp_Renderfarm_File.remove(); //removes our temporary file for the renderfarm
			
		}

		app.endUndoGroup();
	}
	else
		alert("You must have at least 1 render item queued in order to send your project to the renderfarm.");	
}
	
function getFilename(full_filename) {
	var temp_array = full_filename.split(".");
	return temp_array[0];
}

function unescape_spaces(text) {
		return text.replace(/%20/g, " ");
	}

function convertDashToUnderscore(text) {
		return text.replace(/\-/g, "_");
	}

function make_RenderControlFile(filename) {
	
	filename += "_RCF.txt";
	var RCF_text = "After Effects 8.0v1 Render Control File\n" + 
						"max_machines=99\n" + 
						"num_machines=0\n" +
						"init=0\n" +
						"html_init=0\n" +
						"html_name=\"\"\n";
	
	var RCF_file = new File(filename);
	RCF_file.open("w");
	RCF_file.write(RCF_text);
	RCF_file.close();
}

function check_File_Version(full_filename, current_version) {
	
	var return_file_version = new Array();
	return_file_version[0] = full_filename;
	return_file_version[1] = current_version;
		
	var checkFile = new File(full_filename);	
	
		if (checkFile.exists) {
		
			var tmp_filename = full_filename.split("-");
			var tmp_version_number = tmp_filename[1].split(".");
			var version_number = parseInt(tmp_version_number[0])+ 1;

			if (version_number < 10)  //the version number should always be 2 digits so the render files show alphabetically when sorted by name
				version_number = "0" + version_number;
					
			if (version_number > 99) {
				alert ("You can not have more than 99 versions of a file in the renderfarm.\n" +
						"Please delete some After Effects Project Files and their corresponding Render Control Files (RCF.txt)\n" + 
						"from the rendefarm folder");
				return 0;
			}
				
			var tmp_full_filename = tmp_filename[0] + "-" + version_number + ".aep";
			
			return_file_version[0] = tmp_full_filename;			
			return_file_version[1] = version_number;
			
			return check_File_Version(return_file_version[0], return_file_version[1]);
		} else
			return return_file_version;
}

function get_formatted_date() {
	
	var time = new Date();
	var date = time.getDate();
	var month = time.getMonth()+1;
	var year = time.getFullYear();
	var hours = time.getHours();
	var minutes = time.getMinutes();
	var timeofDay;	//am or pm
	if (minutes < 10)
		minutes = "0" + minutes;

	if (hours <= 12)
		timeOfDay = "AM";
	else {
		hours -= 12;
		timeOfDay = "PM";
	}

	return month+"_"+date+"_"+year+"_"+hours+"_"+minutes+"_"+timeOfDay;
}
